git diff HEAD:examples/oauth.ts examples/oidc.ts > examples/oidc.diff
git diff HEAD:examples/oauth.ts examples/par.ts > examples/par.diff
git diff HEAD:examples/oauth.ts examples/dpop.ts > examples/dpop.diff
git diff HEAD:examples/oauth.ts examples/jar.ts > examples/jar.diff
git diff HEAD:examples/oauth.ts examples/jarm.ts > examples/jarm.diff
